/*
 * MPI version with the matrix subdivided by "lines".
 */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <mpi.h>
#include "../config.h"
#include "../utils.h"

#define TAG_BORDER 0
#define TAG_MATRIX 1

float *compute_jacobi(int rank, int numprocs, int n, float init_value, float threshold, borders b, int *iterations) {
  float *complete_x;
  float *x;
  float *new_x;
  float *tmp_x;
  float max_diff, global_max_diff, new_value;
  int i, j;
  int nb = n + 2; // n plus the border
  int rows, rows_to_transmit;
  int receive_pos;

  if (rank == 0) {
    rows = n - (n / numprocs) * (numprocs - 1);
  } else {
    rows = n / numprocs;
  }
  LOG(printf("[Process %d/%d] rows: %d\n", rank, numprocs, rows));
  /* LOG(printf("[Process %d/%d] initializing matrix\n", rank, numprocs)); */
  /* Initialize the matrix */
  x = create_sa_matrix(rows + 2, nb);
  new_x = create_sa_matrix(rows + 2, nb);
  for (i = 0; i < rows + 2; i++) {
    for (j = 1; j <= n; j++) {
      x[IDX(nb, i, j)] = init_value;
      new_x[IDX(nb, i, j)] = init_value;
    }
  }
  /* Initialize boundary regions */
  for (i = 0; i < rows + 2; i++) {
    x[IDX(nb, i, 0)] = b.west;
    x[IDX(nb, i, n + 1)] = b.east;
    new_x[IDX(nb, i, 0)] = b.west;
    new_x[IDX(nb, i, n + 1)] = b.east;
  }
  if (rank == 0) {
    for (i = 1; i <= n + 1; i++) {
      x[IDX(nb, 0, i)] = b.north;
      new_x[IDX(nb, 0, i)] = b.north;
    }
  }
  if (rank == numprocs - 1){
    for (i = 1; i < n + 1; i++) {
      x[IDX(nb, rows + 1, i)] = b.south;
      new_x[IDX(nb, rows + 1, i)] = b.south;
    }
  }
  /* LOG(printf("[Process %d/%d] matrix initialized\n", rank, numprocs)); */
  /* Iterative refinement of x until values converge */
  *iterations = 0;
  do {
    max_diff = 0;
    global_max_diff = 0;
    for (i = 1; i <= rows; i++) {
      for (j = 1; j <= n; j++) {
        new_value = 0.25 * (x[IDX(nb, i - 1, j)] + x[IDX(nb, i, j + 1)] + x[IDX(nb, i + 1, j)] + x[IDX(nb, i, j - 1)]);
        max_diff = fmaxf(max_diff, fabs(new_value - x[IDX(nb, i, j)]));
        new_x[IDX(nb, i, j)] = new_value;
      }
    }
    tmp_x = new_x;
    new_x = x;
    x = tmp_x;
    if (rank % 2 == 0) {
      if (rank != numprocs - 1) {
        // Send and receive south border
        MPI_Send(&x[IDX(nb, rows, 0)], nb, MPI_FLOAT, rank + 1, TAG_BORDER, MPI_COMM_WORLD);
        MPI_Recv(&x[IDX(nb, rows + 1, 0)], nb, MPI_FLOAT, rank + 1, TAG_BORDER, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      }
      if (rank != 0) {
        // Send and receive north border
        MPI_Send(&x[IDX(nb, 1, 0)], nb, MPI_FLOAT, rank - 1, TAG_BORDER, MPI_COMM_WORLD);
        MPI_Recv(&x[IDX(nb, 0, 0)], nb, MPI_FLOAT, rank - 1, TAG_BORDER, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      }
    } else {
      // Receive and send north border
      MPI_Recv(&x[IDX(nb, 0, 0)], nb, MPI_FLOAT, rank - 1, TAG_BORDER, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      MPI_Send(&x[IDX(nb, 1, 0)], nb, MPI_FLOAT, rank - 1, TAG_BORDER, MPI_COMM_WORLD);
      if (rank != numprocs - 1) {
        // Receive and send south border
        MPI_Recv(&x[IDX(nb, rows + 1, 0)], nb, MPI_FLOAT, rank + 1, TAG_BORDER, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        MPI_Send(&x[IDX(nb, rows, 0)], nb, MPI_FLOAT, rank + 1, TAG_BORDER, MPI_COMM_WORLD);
      }
    }
    /* LOG(printf("[Process %d/%d] max_diff: %f\n", rank, numprocs, max_diff)); */
    MPI_Allreduce(&max_diff, &global_max_diff, 1, MPI_FLOAT, MPI_MAX, MPI_COMM_WORLD);
    /* LOG(printf("[Process %d/%d] global_max_diff: %f\n", rank, numprocs, global_max_diff)); */
    (*iterations)++;
  } while (global_max_diff > threshold);

  if (rank == 0) {
    complete_x = create_sa_matrix(nb, nb);
    memcpy(complete_x, x, (rows + ((rank == numprocs - 1) ? 2 : 1)) * (nb) * sizeof(float));
    rows_to_transmit = n / numprocs;
    receive_pos = rows + 1;
    for (i = 1; i < numprocs; i++) {
      if (i == numprocs - 1) {
        rows_to_transmit++;
      }
      MPI_Recv(&complete_x[IDX(nb, receive_pos, 0)], rows_to_transmit * (nb), MPI_FLOAT, i, TAG_MATRIX, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      receive_pos += n / numprocs;
    }
  } else {
    complete_x = NULL;
    rows_to_transmit = rows;
    if (rank == numprocs - 1) {
      rows_to_transmit++;
    }
    MPI_Send(&x[IDX(nb, 1, 0)], rows_to_transmit * (nb), MPI_FLOAT, 0, TAG_MATRIX, MPI_COMM_WORLD);
  }

  return complete_x;
}
